from bedrock_protocol.packets.packet import BlockActorDataPacket, ContainerOpenPacket
from bedrock_protocol.packets.types import BlockPos, NetworkBlockPosition
from endstone import Player
from rapidnbt import CompoundTag

from endstone_inventoryui.menu import Menu
from endstone_inventoryui.menu.graphic.graphic import Graphic
from endstone_inventoryui.util.utils import east, west, send_block


class BlockPairGraphic(Graphic):
    menu: Menu
    pos: tuple[BlockPos, BlockPos]

    def __init__(self, menu: Menu, pos: BlockPos):
        self.menu = menu
        if (pos.x & 1) == 1:
            self.pos = (pos, east(pos))
        else:
            self.pos = (pos, west(pos))

    def send(self, player: Player):
        for p in self.pos:
            send_block(player, self.menu.type.get_block_id(), p)

    def send_data(self, player: Player):
        for p in self.pos:
            tag = CompoundTag()
            tag.set("id", self.menu.type.get_block_actor_id())
            tag.set("CustomName", self.menu.name)
            pk = BlockActorDataPacket(NetworkBlockPosition(p.x, p.y, p.z), tag)
            player.send_packet(pk.get_packet_id(), pk.serialize())

    def open(self, player: Player):
        from endstone_inventoryui.manager import Session
        pos, _ = self.pos
        net_pos = NetworkBlockPosition(pos.x, pos.y, pos.z)
        pk = ContainerOpenPacket(Session.CONTAINER_ID, self.menu.type.get_container_type(), net_pos)
        player.send_packet(pk.get_packet_id(), pk.serialize())

    def remove(self, player: Player):
        for p in self.pos:
            block = player.dimension.get_block_at(p.x, p.y, p.z)
            send_block(player, block.type, p)



"""
NBT Network Little Endian Serializer for Minecraft Bedrock Edition
"""

from typing import Optional
import struct

from bstream import BinaryStream

# Tag type IDs for NBT format
TAG_END = 0
TAG_BYTE = 1
TAG_SHORT = 2
TAG_INT = 3
TAG_LONG = 4
TAG_FLOAT = 5
TAG_DOUBLE = 6
TAG_BYTE_ARRAY = 7
TAG_STRING = 8
TAG_LIST = 9
TAG_COMPOUND = 10
TAG_INT_ARRAY = 11
TAG_LONG_ARRAY = 12


class NBTSerializer:
    """Serializes NBT tags to network little endian format for Bedrock Edition"""

    def __init__(self, stream):
        """
        Initialize the serializer with a BinaryStream

        Args:
            stream: BinaryStream instance (should be in little endian mode)
        """
        self.stream: BinaryStream = stream

    def write_tag(self, tag, name: Optional[str] = None, write_type: bool = True):
        """
        Write a complete NBT tag with optional name and type

        Args:
            tag: The NBT tag to serialize
            name: Optional tag name (for named tags)
            write_type: Whether to write the tag type ID
        """
        tag_type = self._get_tag_type(tag)

        if write_type:
            self.stream.write_byte(tag_type)

        if name is not None and tag_type != TAG_END:
            self._write_string(name)

        self._write_tag_payload(tag, tag_type)

    def _get_tag_type(self, tag) -> int:
        """Determine the NBT type ID for a tag"""
        type_name = type(tag).__name__

        type_map = {
            'ByteTag': TAG_BYTE,
            'ShortTag': TAG_SHORT,
            'IntTag': TAG_INT,
            'LongTag': TAG_LONG,
            'FloatTag': TAG_FLOAT,
            'DoubleTag': TAG_DOUBLE,
            'ByteArrayTag': TAG_BYTE_ARRAY,
            'StringTag': TAG_STRING,
            'ListTag': TAG_LIST,
            'CompoundTag': TAG_COMPOUND,
            'IntArrayTag': TAG_INT_ARRAY,
        }

        return type_map.get(type_name, TAG_END)

    def _write_string(self, value: str):
        """Write a string with length prefix (unsigned short)"""
        encoded = value.encode('utf-8')
        self.stream.write_unsigned_short(len(encoded))
        self.stream.write_raw_bytes(encoded)

    def _write_tag_payload(self, tag, tag_type: int):
        """Write the payload data for a tag based on its type"""
        if tag_type == TAG_END:
            return
        elif tag_type == TAG_BYTE:
            self.stream.write_byte(tag.value)
        elif tag_type == TAG_SHORT:
            self.stream.write_signed_short(tag.value)
        elif tag_type == TAG_INT:
            self.stream.write_signed_int(tag.value)
        elif tag_type == TAG_LONG:
            self.stream.write_signed_int64(tag.value)
        elif tag_type == TAG_FLOAT:
            self.stream.write_float(tag.value)
        elif tag_type == TAG_DOUBLE:
            self.stream.write_double(tag.value)
        elif tag_type == TAG_BYTE_ARRAY:
            self._write_byte_array(tag)
        elif tag_type == TAG_STRING:
            self._write_string(tag.value)
        elif tag_type == TAG_LIST:
            self._write_list(tag)
        elif tag_type == TAG_COMPOUND:
            self._write_compound(tag)
        elif tag_type == TAG_INT_ARRAY:
            self._write_int_array(tag)

    def _write_byte_array(self, tag):
        """Write a byte array tag"""
        self.stream.write_signed_int(len(tag))
        for byte_val in tag:
            self.stream.write_byte(byte_val)

    def _write_int_array(self, tag):
        """Write an int array tag"""
        self.stream.write_signed_int(len(tag))
        for int_val in tag:
            self.stream.write_signed_int(int_val)

    def _write_list(self, tag):
        """Write a list tag"""
        if len(tag) == 0:
            # Empty list - use TAG_END as element type
            self.stream.write_byte(TAG_END)
            self.stream.write_signed_int(0)
        else:
            # Get type from first element
            element_type = self._get_tag_type(tag[0])
            self.stream.write_byte(element_type)
            self.stream.write_signed_int(len(tag))

            # Write all elements without type/name prefixes
            for element in tag:
                self._write_tag_payload(element, element_type)

    def _write_compound(self, tag):
        """Write a compound tag"""
        for key in tag.keys():
            value = tag[key]
            self.write_tag(value, name=key, write_type=True)

        # Write TAG_END to mark end of compound
        self.stream.write_byte(TAG_END)

    def serialize_root(self, tag, root_name: str = ""):
        """
        Serialize a root NBT tag (typically a CompoundTag)

        Args:
            tag: The root NBT tag
            root_name: Name for the root tag (default: empty string)
        """
        self.write_tag(tag, name=root_name, write_type=True)


# Convenience function
def write_nbt(stream, tag, root_name: str = "") -> bytes:
    """
    Write an NBT tag to a BinaryStream

    Args:
        stream: BinaryStream instance
        tag: NBT tag to serialize
        root_name: Optional root tag name

    Returns:
        The serialized data from the stream
    """
    serializer = NBTSerializer(stream)
    serializer.serialize_root(tag, root_name)
    return stream.data()

#def serialize(tag: CompoundTag)


from enum import Enum


class MenuType(Enum):
    CHEST = 0
    DOUBLE_CHEST = 1
    HOPPER = 2

    def is_pair(self) -> bool:
        match self:
            case MenuType.CHEST:
                return False
            case MenuType.DOUBLE_CHEST:
                return True
            case MenuType.HOPPER:
                return False

    def get_block_id(self) -> str:
        match self:
            case MenuType.CHEST | MenuType.DOUBLE_CHEST:
                return "minecraft:chest"
            case MenuType.HOPPER:
                return "minecraft:hopper"

    def get_container_type(self) -> int:
        match self:
            case MenuType.CHEST | MenuType.DOUBLE_CHEST:
                return 0x0
            case MenuType.HOPPER:
                return 0x8

    def get_block_actor_id(self):
        match self:
            case MenuType.CHEST | MenuType.DOUBLE_CHEST:
                return "Chest"
            case MenuType.HOPPER:
                return "Hopper"

    def get_container_size(self) -> int:
        match self:
            case MenuType.CHEST:
                return 27
            case MenuType.DOUBLE_CHEST:
                return 54
            case MenuType.HOPPER:
                return 5



from endstone import Player
from endstone._internal.endstone_python import ItemStack
from endstone.command import CommandSender, Command
from endstone.plugin import Plugin

from endstone_inventoryui import *


class Main(Plugin):
    prefix = "InventoryUI Example"
    api_version = "0.10"
    load = "POSTWORLD"
    commands = {
        "chestmenu": {
            "description": "Opens chest menu",
            "usages": ["/chestmenu"],
        }
    }

    chest_menu = Menu(MenuType.DOUBLE_CHEST, "chest menu")

    def on_command(self, sender: CommandSender, command: Command, _):
        match command.name:
            case "chestmenu":
                if isinstance(sender, Player):
                    inventory = self.chest_menu.inventory
                    self.chest_menu.set_listener(self.on_click)

                    inventory.add_item(ItemStack("minecraft:diamond_sword"))
                    inventory.add_item(ItemStack("minecraft:diamond_axe"))
                    inventory.add_item(ItemStack("minecraft:diamond_pickaxe"))
                    inventory.add_item(ItemStack("minecraft:hopper"))

                    ench_item = ItemStack("minecraft:diamond_sword")
                    meta = ench_item.item_meta
                    meta.add_enchant("sharpness", 5)
                    ench_item.set_item_meta(meta)
                    inventory.set_item(5, ench_item)

                    self.chest_menu.send_to(sender)
        return True

    def on_click(self, player: Player, slot: int, item: ItemStack, inventory: UIInventory):
        print(f"{player.name} clicked on slot {slot} ({item.type.id})")
        if slot == 3:
            menu = Menu(MenuType.HOPPER, "hopper menu")
            menu.inventory.set_item(0, ItemStack("minecraft:diamond_sword"))
            menu.inventory.set_item(1, ItemStack("minecraft:diamond_axe"))
            menu.inventory.set_item(2, ItemStack("minecraft:diamond_pickaxe"))
            self.chest_menu.close(player)
            menu.send_to(player)
            return



from endstone.plugin import Plugin

from endstone_inventoryui.listener import EventListener


class InventoryUIPlugin(Plugin):
    prefix = "InventoryUI"
    api_version = "0.6"
    load = "POSTWORLD"

    def on_enable(self) -> None:
        self.register_events(EventListener(self))



from typing import TYPE_CHECKING
from uuid import UUID

from endstone import Player

if TYPE_CHECKING:
    from endstone_inventoryui.manager import Session

sessions: dict[UUID, 'Session'] = {}


def find_session(player: Player) -> 'Session|None':
    global sessions
    if player.unique_id in sessions:
        return sessions[player.unique_id]
    return None


def create_session(player: Player) -> 'Session':
    from endstone_inventoryui.manager import Session  # Import here at runtime

    global sessions
    session = Session(player=player)
    sessions[player.unique_id] = session
    return session


def close_session(player: Player):
    global sessions
    if player.unique_id in sessions:
        del sessions[player.unique_id]


def get_all_sessions():
    global sessions
    return sessions.values()



from bedrock_protocol.packets.types import ItemData
from bstream import BinaryStream
from endstone._internal.endstone_python import ItemStack

from endstone_inventoryui.util.item_utils import build_tag
from endstone_inventoryui.util.item_utils import is_air


class ItemStackWrapper:
    stack_id: int
    item_stack: ItemStack
    data: ItemData

    def __init__(self, stack_id: int = 0, item_stack: ItemStack | None = None):
        from endstone_inventoryui.util.item_utils import get_item_data
        self.stack_id: int = stack_id
        self.item_stack: ItemStack = item_stack or ItemStack("minecraft:air")
        data = get_item_data(self.item_stack.type.id)
        if data is None:
            raise ValueError(f"ItemStackWrapper: ItemData not found for {item_stack.type.id}")
        self.data: ItemData = data

    def write_header(self, stream: BinaryStream) -> bool:
        if is_air(self.item_stack):
            stream.write_varint(0)
            return False
        stream.write_varint(self.data.item_id)
        stream.write_unsigned_short(self.item_stack.amount)
        stream.write_unsigned_varint(self.item_stack.data)
        return True

    def write_footer(self, stream: BinaryStream):
        """Builds the extra data buffer (NBT + canPlaceOn + canDestroy)"""
        item_meta = self.item_stack.item_meta
        tag = build_tag(item_meta)
        if not tag.empty():
            stream.write_signed_short(-1)  # nbt length
            stream.write_byte(1)  # nbt version?
            stream.write_raw_bytes(tag.to_binary_nbt())
        else:
            stream.write_signed_short(0)  # no nbt

        stream.write_unsigned_int(0)  # canPlaceOn count
        stream.write_unsigned_int(0)  # canDestroy count

    def write(self, stream: BinaryStream):
        if self.write_header(stream):
            has_net_id = self.stack_id != 0
            stream.write_bool(has_net_id)
            if has_net_id:
                stream.write_varint(self.stack_id)

            stream.write_varint(0)  # BlockRuntimeID

            user_data = BinaryStream()
            self.write_footer(user_data)
            stream.write_bytes(user_data.copy_buffer())  # user data
